#ifndef DUNE_QUADRATIC_IPOPT_SOLVER_HH
#define DUNE_QUADRATIC_IPOPT_SOLVER_HH

/** \file
    \brief A wrapper for the IPOpt solver for quadratic optimization problem
    with box constraints.
*/
#include <algorithm>
#include <vector>

#include <dune/common/exceptions.hh>
#include <dune/common/bitsetvector.hh>
#include <dune/common/fmatrix.hh>

#include <dune/istl/bcrsmatrix.hh>
#include <dune/istl/bvector.hh>

#include <dune/solvers/common/boxconstraint.hh>
#include <dune/solvers/common/canignore.hh>
#include <dune/solvers/solvers/iterativesolver.hh>

#include "IpTNLP.hpp"
#include "IpIpoptApplication.hpp"

/** \brief Implementation class used to pipe quadratic problems to
    the IPOpt interior-point solver
*/
template <class MatrixType, class VectorType>
class QuadraticIPOptProblem : public Ipopt::TNLP
{
    enum {blocksize = VectorType::block_type::dimension};

    typedef typename VectorType::field_type field_type;

public:
  /** default constructor */
    QuadraticIPOptProblem(const MatrixType* hessian, VectorType* x, const VectorType* rhs,
                          const Dune::BitSetVector<blocksize>* dirichletNodes,
                          std::vector<BoxConstraint<field_type, blocksize> >* obstacles)
        : hessian_(hessian), x_(x), rhs_(rhs), 
          dirichletNodes_(dirichletNodes), obstacles_(obstacles)
    {}

  /** default destructor */
    virtual ~QuadraticIPOptProblem() {};

  /**@name Overloaded from TNLP */
  //@{
  /** Method to return some info about the nlp */
  virtual bool get_nlp_info(Ipopt::Index& n, Ipopt::Index& m, Ipopt::Index& nnz_jac_g,
                            Ipopt::Index& nnz_h_lag, IndexStyleEnum& index_style);

  /** Method to return the bounds for my problem */
  virtual bool get_bounds_info(Ipopt::Index n, Ipopt::Number* x_l, Ipopt::Number* x_u,
                               Ipopt::Index m, Ipopt::Number* g_l, Ipopt::Number* g_u);

  /** Method to return the starting point for the algorithm */
  virtual bool get_starting_point(Ipopt::Index n, bool init_x, Ipopt::Number* x,
                                  bool init_z, Ipopt::Number* z_L, Ipopt::Number* z_U,
                                  Ipopt::Index m, bool init_lambda,
                                  Ipopt::Number* lambda);

  /** Method to return the objective value */
  virtual bool eval_f(Ipopt::Index n, const Ipopt::Number* x, bool new_x, Ipopt::Number& obj_value);

  /** Method to return the gradient of the objective */
  virtual bool eval_grad_f(Ipopt::Index n, const Ipopt::Number* x, bool new_x, Ipopt::Number* grad_f);

  /** Method to return the constraint residuals */
  virtual bool eval_g(Ipopt::Index n, const Ipopt::Number* x, bool new_x, Ipopt::Index m, Ipopt::Number* g);

  /** Method to return:
   *   1) The structure of the jacobian (if "values" is NULL)
   *   2) The values of the jacobian (if "values" is not NULL)
   */
  virtual bool eval_jac_g(Ipopt::Index n, const Ipopt::Number* x, bool new_x,
                          Ipopt::Index m, Ipopt::Index nele_jac, Ipopt::Index* iRow, Ipopt::Index *jCol,
                          Ipopt::Number* values);

  /** Method to return:
   *   1) The structure of the hessian of the lagrangian (if "values" is NULL)
   *   2) The values of the hessian of the lagrangian (if "values" is not NULL)
   */
  virtual bool eval_h(Ipopt::Index n, const Ipopt::Number* x, bool new_x,
                      Ipopt::Number obj_factor, Ipopt::Index m, const Ipopt::Number* lambda,
                      bool new_lambda, Ipopt::Index nele_hess, Ipopt::Index* iRow,
                      Ipopt::Index* jCol, Ipopt::Number* values);

  //@}

  /** @name Solution Methods */
  //@{
  /** This method is called when the algorithm is complete so the TNLP can store/write the solution */
    virtual void finalize_solution(Ipopt::SolverReturn status,
                                 Ipopt::Index n, const Ipopt::Number* x, const Ipopt::Number* z_L, const Ipopt::Number* z_U,
                                 Ipopt::Index m, const Ipopt::Number* g, const Ipopt::Number* lambda,
                                 Ipopt::Number obj_value,
                                   const Ipopt::IpoptData* ip_data,
                                   Ipopt::IpoptCalculatedQuantities* ip_cq);
  //@}

    // /////////////////////////////////
    //   Data
    // /////////////////////////////////

    const MatrixType* hessian_;

    VectorType* x_;

    const VectorType* rhs_;

    const Dune::BitSetVector<blocksize>* dirichletNodes_;

    std::vector<BoxConstraint<field_type, blocksize> >* obstacles_;

private:
  /**@name Methods to block default compiler methods.
   */
  //@{
  //  QuadraticIPOptProblem();
  QuadraticIPOptProblem(const QuadraticIPOptProblem&);
  QuadraticIPOptProblem& operator=(const QuadraticIPOptProblem&);
  //@}
};

// returns the size of the problem
template <class MatrixType, class VectorType>
bool QuadraticIPOptProblem<MatrixType,VectorType>::
get_nlp_info(Ipopt::Index& n, Ipopt::Index& m, Ipopt::Index& nnz_jac_g,
             Ipopt::Index& nnz_h_lag, IndexStyleEnum& index_style)
{
    // The problem described in HS071_NLP.hpp has 4 variables, x[0] through x[3]
    n = x_->dim();

    // No equality constraints: Dirichlet conditions are actually inequality constraints
    m = 0;

    // hence the constraint jacobian is empty
    nnz_jac_g = 0;

    // We only need the lower left corner of the Hessian (since it is symmetric)
    if (!hessian_)
        DUNE_THROW(SolverError, "No Hessian matrix has been supplied!");

    nnz_h_lag = 0;
    
    for (size_t rowIdx=0; rowIdx<hessian_->N(); rowIdx++) {
        
        const typename MatrixType::row_type& row = (*hessian_)[rowIdx];
        
        typename MatrixType::row_type::ConstIterator cIt    = row.begin();
        typename MatrixType::row_type::ConstIterator cEndIt = row.end();
        
        for (; cIt!=cEndIt; ++cIt) {
            
            if (rowIdx > cIt.index()) {

                nnz_h_lag += blocksize*blocksize;
                
            } else if (rowIdx == cIt.index()) {

                nnz_h_lag += ((blocksize+1)*blocksize) / 2;
                
            }
            
        }
        
    }
    
    // use the C style indexing (0-based)
    index_style = Ipopt::TNLP::C_STYLE;
    
    return true;
}


// returns the variable bounds
template <class MatrixType, class VectorType>
bool QuadraticIPOptProblem<MatrixType,VectorType>::
get_bounds_info(Ipopt::Index n, Ipopt::Number* x_l, Ipopt::Number* x_u,
                Ipopt::Index m, Ipopt::Number* g_l, Ipopt::Number* g_u)
{
    // here, the n and m we gave IPOPT in get_nlp_info are passed back to us.
    // If desired, we could assert to make sure they are what we think they are.
    assert(n == (Ipopt::Index)x_->dim());
    assert(m == 0);

    if (obstacles_) {

        for (size_t i=0; i<x_->size(); i++) {
            
            for (size_t j=0; j<blocksize; j++) {
                
                // Simply copy the values.  If they exceed a certain limit they are
                // considered 'off'.
                x_l[i*blocksize+j] = (*obstacles_)[i].lower(j);
                x_u[i*blocksize+j] = (*obstacles_)[i].upper(j);
                
            }
            
        }

    } else {

        // unset all bounds
        for (size_t i=0; i<x_->dim(); i++) {
            x_l[i] = -std::numeric_limits<double>::max();
            x_u[i] =  std::numeric_limits<double>::max();
        }
        
    }

    if (dirichletNodes_) {

        for (size_t i=0; i<x_->size(); i++) {
            
            for (size_t j=0; j<blocksize; j++) {

                if ( (*dirichletNodes_)[i][j])
                    x_l[i*blocksize+j] = x_u[i*blocksize+j] = (*this->x_)[i][j];
                
            }
            
        }

    }

  return true;
}

// returns the initial point for the problem
template <class MatrixType, class VectorType>
bool QuadraticIPOptProblem<MatrixType,VectorType>::
get_starting_point(Ipopt::Index n, bool init_x, Ipopt::Number* x,
                   bool init_z, Ipopt::Number* z_L, Ipopt::Number* z_U,
                   Ipopt::Index m, bool init_lambda, Ipopt::Number* lambda)
{
    // Here, we assume we only have starting values for x, if you code
    // your own NLP, you can provide starting values for the dual variables
    // if you wish
    assert(init_x == true);
    assert(init_z == false);
    assert(init_lambda == false);
    
    // initialize to the given starting point
    for (size_t i=0; i<x_->size(); i++)
        for (int j=0; j<blocksize; j++)
            x[i*blocksize+j] = (*x_)[i][j];

    return true;
}

// returns the value of the objective function
template <class MatrixType, class VectorType>
bool QuadraticIPOptProblem<MatrixType,VectorType>::
eval_f(Ipopt::Index n, const Ipopt::Number* x, bool new_x, Ipopt::Number& obj_value)
{
//     std::cout << "x:" << std::endl;
//     for (int i=0; i<n; i++)
//         std::cout << x[i] << std::endl;

    // Init return value
    obj_value = 0;

    ////////////////////////////////////
    // Compute x^T*A*x
    ////////////////////////////////////
    
    for (size_t rowIdx=0; rowIdx<hessian_->N(); rowIdx++) {
        
        const typename MatrixType::row_type& row = (*hessian_)[rowIdx];
        
        typename MatrixType::row_type::ConstIterator cIt   = row.begin();
        typename MatrixType::row_type::ConstIterator cEndIt = row.end();
        
        for (; cIt!=cEndIt; ++cIt) {

            // Multiply block
            for (int i=0; i<blocksize; i++)
                for (int j=0; j<blocksize; j++)
                    obj_value += x[rowIdx*blocksize+i] * x[cIt.index()*blocksize+j] * (*cIt)[i][j];

        }

    }

    //std::cout << "bilinear form: " << *f << std::endl;
    
    // because J = 1/2 * A(x,x) - b(x)
    obj_value *= 0.5;
    
    // -b(x)
    for (size_t i=0; i<rhs_->size(); i++)
        for (int j=0; j<blocksize; j++)
            obj_value -= x[i*blocksize+j] * (*rhs_)[i][j];
  
    //std::cout << "IPOPT Energy: " << obj_value << std::endl;
    //exit(0);

  return true;
}

// return the gradient of the objective function grad_{x} f(x)
template <class MatrixType, class VectorType>
bool QuadraticIPOptProblem<MatrixType,VectorType>::
eval_grad_f(Ipopt::Index n, const Ipopt::Number* x, bool new_x, Ipopt::Number* grad_f)
{
    //std::cout << "### eval_grad_f ###" << std::endl;

    // \nabla J = A(x,.) - b(x)
    for (int i=0; i<n; i++)
        grad_f[i] = 0;

    for (size_t i=0; i<hessian_->N(); i++) {

        const typename MatrixType::row_type& row = (*hessian_)[i];

        typename MatrixType::row_type::ConstIterator cIt   = row.begin();
        typename MatrixType::row_type::ConstIterator cEndIt = row.end();
        
        for (; cIt!=cEndIt; ++cIt) {
            
            for (int j=0; j<blocksize; j++)
                for (int k=0; k<blocksize; k++)
                    grad_f[i*blocksize+j] += (*cIt)[j][k] * x[cIt.index()*blocksize+k];

        }

    }
    
    // g = g - b
    for (size_t i=0; i<rhs_->size(); i++)
        for (int j=0; j<blocksize; j++)
            grad_f[i*blocksize+j] -= (*rhs_)[i][j];

//     for (int i=0; i<n; i++)
//         printf("%d:  x = %g,  rhs = %g,   g = %g\n", i, x[i], (*rhs_)[i/blocksize][i%blocksize], grad_f[i]);
  return true;
}

// return the value of the constraints: g(x)
template <class MatrixType, class VectorType>
bool QuadraticIPOptProblem<MatrixType,VectorType>::
eval_g(Ipopt::Index n, const Ipopt::Number* x, bool new_x, Ipopt::Index m, Ipopt::Number* g)
{
  return true;
}

// return the structure or values of the jacobian
template <class MatrixType, class VectorType>
bool QuadraticIPOptProblem<MatrixType,VectorType>::
eval_jac_g(Ipopt::Index n, const Ipopt::Number* x, bool new_x,
           Ipopt::Index m, Ipopt::Index nele_jac, Ipopt::Index* iRow, Ipopt::Index *jCol,
           Ipopt::Number* values)
{
  return true;
}

//return the structure or values of the hessian
template <class MatrixType, class VectorType>
bool QuadraticIPOptProblem<MatrixType,VectorType>::
eval_h(Ipopt::Index n, const Ipopt::Number* x, bool new_x,
       Ipopt::Number obj_factor, Ipopt::Index m, const Ipopt::Number* lambda,
       bool new_lambda, Ipopt::Index nele_hess, Ipopt::Index* iRow,
       Ipopt::Index* jCol, Ipopt::Number* values)
{
    if (values == NULL) {

        // return the structure. This is a symmetric matrix, fill the lower left
        // triangle only.
        
        int idx = 0;
        
        // Loop over all matrix rows
        for (size_t rowIdx=0; rowIdx<hessian_->N(); rowIdx++) {
            
            const typename MatrixType::row_type& row = (*hessian_)[rowIdx];
            
            typename MatrixType::row_type::ConstIterator cIt   = row.begin();
            typename MatrixType::row_type::ConstIterator cEndIt = row.end();
            
            // Loop over all columns in this row
            for (; cIt!=cEndIt; ++cIt) {
                
                // Current entry on the lower left triangular matrix and not on the diagonal
                // --> copy it completely
                if (rowIdx > cIt.index()) {
                    
                    for (int i=0; i<blocksize; i++) {
                        
                        for (int j=0; j<blocksize; j++) {
                            
                            iRow[idx] = rowIdx*blocksize + i;
                            jCol[idx] = cIt.index()*blocksize + j;
                            idx++;
                            
                        }
                        
                    }
                    
                } else if (rowIdx == cIt.index()) {
                    
                    // Current entry is on the diagonal
                    // --> copy only its lower triangular half
                    for (int i=0; i<blocksize; i++) {
                        
                        for (int j=0; j<i+1; j++) {
                            
                            iRow[idx] = rowIdx*blocksize + i;
                            jCol[idx] = cIt.index()*blocksize + j;
                            idx++;
                            
                        }
                        
                    }
                    
                }
                
            }
            
        }
        
        assert(idx == nele_hess);
//         for (int i=0; i<nele_hess; i++)
//             printf("%d %d \n", iRow[i], jCol[i]);

    } else {
        
        // return the values. This is a symmetric matrix, fill the lower left
        // triangle only
        
        int idx = 0;
        
        // Loop over all matrix rows
        for (size_t rowIdx=0; rowIdx<hessian_->N(); rowIdx++) {
            
            const typename MatrixType::row_type& row = (*hessian_)[rowIdx];
            
            typedef typename MatrixType::row_type::ConstIterator ColumnIterator;
            
            ColumnIterator cIt   = row.begin();
            ColumnIterator cEndIt = row.end();
            
            // Loop over all columns in this row
            for (; cIt!=cEndIt; ++cIt) {
                
                // Current entry on the lower left triangular matrix and not on the diagonal
                // --> copy it completely
                if (rowIdx > cIt.index()) {
                    
                    for (int i=0; i<blocksize; i++)
                        for (int j=0; j<blocksize; j++) {
                            //printf("%d %d %g\n", rowIdx*blocksize + i, cIt.index()*blocksize+j,(*cIt)[i][j]);
                            values[idx++] = obj_factor * (*cIt)[i][j];
                        }
                    
                } else if (rowIdx == cIt.index()) {
                    
                    // Current entry is on the diagonal
                    // --> copy only its lower triangular half
                    for (int i=0; i<blocksize; i++)
                        for (int j=0; j<i+1; j++) {
                            //printf("%d %d %g\n", rowIdx*blocksize + i, cIt.index()*blocksize+j,(*cIt)[i][j]);
                            values[idx++] = obj_factor * (*cIt)[i][j];
                        }
                }
                
            }
            
        }

        assert(idx == nele_hess);

//         for (int i=0; i<nele_hess; i++)
//             printf("%g\n", values[i]);

    }

    return true;
}

template <class MatrixType, class VectorType>
void QuadraticIPOptProblem<MatrixType,VectorType>::
finalize_solution(Ipopt::SolverReturn status,
                  Ipopt::Index n, const Ipopt::Number* x, const Ipopt::Number* z_L, const Ipopt::Number* z_U,
                  Ipopt::Index m, const Ipopt::Number* g, const Ipopt::Number* lambda,
                  Ipopt::Number obj_value,
                  const Ipopt::IpoptData* ip_data,
                  Ipopt::IpoptCalculatedQuantities* ip_cq)
{
  // here is where we would store the solution to variables, or write to a file, etc
  // so we could use the solution.

    for (size_t i=0; i<x_->size(); i++)
        for (int j=0; j<blocksize; j++)
            (*x_)[i][j] = x[i*blocksize+j];

}


/** \brief Wraps the IPOpt interior-point solver for quadratic problems with
    bound constraints.
*/
template <class MatrixType, class VectorType>
class QuadraticIPOptSolver 
    : public IterativeSolver<VectorType>, public CanIgnore<Dune::BitSetVector<VectorType::block_type::dimension> >
{

    enum {blocksize = VectorType::block_type::dimension};

    typedef typename VectorType::field_type field_type;

    typedef Dune::FieldMatrix<field_type, blocksize, blocksize> MatrixBlock;
    typedef Dune::FieldVector<field_type, blocksize>            VectorBlock;


public:

    /** \brief Default constructor */
    QuadraticIPOptSolver () : IterativeSolver<VectorType>(1e-8, 100, NumProc::FULL),
                              hessian_(NULL), rhs_(NULL),
                              obstacles_(NULL)
    {}

    /** \brief Constructor for a linear problem */
    QuadraticIPOptSolver (const MatrixType& hessian, 
                       VectorType& x,
                       const VectorType& rhs)
        : IterativeSolver<VectorType>(1e-8, 100, NumProc::FULL),
          hessian_(&hessian), rhs_(&rhs), obstacles_(NULL)
    {
        this->x_ = &x;
    }

    void setProblem(const MatrixType& hessian, 
                    VectorType& x,
                    const VectorType& rhs) {
        hessian_ = &hessian;
        x_ = &x;
        rhs_ = &rhs;
    }

    virtual void solve();

    ///////////////////////////////////////////////////////
    // Data
    ///////////////////////////////////////////////////////

    const MatrixType* hessian_;

    VectorType* x_;

    const VectorType* rhs_;

    std::vector<BoxConstraint<field_type,blocksize> >* obstacles_;

};

template <class MatrixType, class VectorType>
void QuadraticIPOptSolver<MatrixType,VectorType>::solve()
{
  // Create a new instance of your nlp
  //  (use a SmartPtr, not raw)
    Ipopt::SmartPtr<Ipopt::TNLP> mynlp 
        = new QuadraticIPOptProblem<MatrixType,VectorType>(hessian_, x_, rhs_,
                                                           this->ignoreNodes_, obstacles_);

  // Create a new instance of IpoptApplication
  //  (use a SmartPtr, not raw)
    Ipopt::SmartPtr<Ipopt::IpoptApplication> app = new Ipopt::IpoptApplication();

  // Change some options
  app->Options()->SetNumericValue("tol", this->tolerance_);
  app->Options()->SetIntegerValue("max_iter", this->maxIterations_);
  app->Options()->SetStringValue("mu_strategy", "adaptive");
  app->Options()->SetStringValue("output_file", "ipopt.out");
  switch (this->verbosity_) {
  case NumProc::QUIET:
      app->Options()->SetIntegerValue("print_level", 0);
      break;
  case NumProc::REDUCED:
      app->Options()->SetIntegerValue("print_level", 5);
      break;
  case NumProc::FULL:
      app->Options()->SetIntegerValue("print_level", 12);
  }

  // Intialize the IpoptApplication and process the options
  Ipopt::ApplicationReturnStatus status;
  status = app->Initialize();
  if (status != Ipopt::Solve_Succeeded)
      DUNE_THROW(Dune::Exception, "IPOpt: Error during initialization!");

  // Ask Ipopt to solve the problem
  status = app->OptimizeTNLP(mynlp);

  // As the SmartPtrs go out of scope, the reference count
  // will be decremented and the objects will automatically
  // be deleted.

}

#endif
